@@ -142,6 +142,13 @@ PerlInterpreter *p5_init_perl(
 
 void p5_destruct_perl(PerlInterpreter *my_perl) {
     PERL_SET_CONTEXT(my_perl);
+
+    SV **cbs_entry = hv_fetchs(PL_modglobal, "Inline::Perl5 callbacks", 0);
+    perl6_callbacks *cbs = NULL;
+    if (cbs_entry) {
+        cbs = (perl6_callbacks*)SvIV(*cbs_entry);
+    }
+
     PL_perl_destruct_level = 1;
 
     POPSTACK_TO(PL_mainstack);
@@ -151,6 +158,9 @@ void p5_destruct_perl(PerlInterpreter *my_perl) {
     perl_destruct(my_perl);
     perl_free(my_perl);
 
+    if (cbs)
+        free(cbs);
+
     if (--interpreters == 0 && terminate)
         PERL_SYS_TERM();
 }
@@ -500,11 +510,15 @@ SV *p5_get_global(PerlInterpreter *my_perl, const char* name) {
     if (name[0] == '$')
         return get_sv(&name[1], 0);
 
-    if (name[0] == '@')
-        return sv_2mortal(newRV_inc((SV *)get_av(&name[1], 0)));
+    if (name[0] == '@') {
+        AV *av = get_av(&name[1], 0);
+        return av ? sv_2mortal(newRV_inc((SV *)av)) : NULL;
+    }
 
-    if (name[0] == '%')
-        return sv_2mortal(newRV_inc((SV *)get_hv(&name[1], 0)));
+    if (name[0] == '%') {
+        HV *hv = get_hv(&name[1], 0);
+        return hv ? sv_2mortal(newRV_inc((SV *)hv)) : NULL;
+    }
 
     return NULL;
 }
@@ -623,7 +637,7 @@ GV *p5_look_up_package_method(PerlInterpreter *my_perl, char *module, char *name
     PERL_SET_CONTEXT(my_perl);
     {
         HV * const pkg = gv_stashpvn(module, strlen(module), 0);
-        GV * const gv = gv_fetchmeth_pvn(pkg, name, strlen(name), -1, SVf_UTF8);
+        GV * const gv = gv_fetchmeth_pvn_autoload(pkg, name, strlen(name), -1, SVf_UTF8);
         if (gv && isGV(gv))
             return gv;
         return NULL;
@@ -635,10 +649,12 @@ SV *p5_call_inherited_package_method(PerlInterpreter *my_perl, char *package, ch
     {
         dSP;
         SV * retval = NULL;
-        HV * stash = gv_stashpvn(package, strlen(package), SVf_UTF8);
+        SV * package_sv = newSVpv(package, 0);
+        HV * stash = gv_stashsv(package_sv, SVf_UTF8);
         int flags = G_ARRAY | G_EVAL;
 
         if (stash == NULL) {
+            SvREFCNT_dec(package_sv);
             *type = -1; /* signal that a wrapper package needs to be created */
             return NULL;
         }
@@ -648,14 +664,12 @@ SV *p5_call_inherited_package_method(PerlInterpreter *my_perl, char *package, ch
 
         PUSHMARK(SP);
 
-        XPUSHs(newSVpv(package, 0));
+        XPUSHs(sv_2mortal(package_sv));
         push_arguments(sp, len, args);
 
-        {
-            GV * const gv = p5_look_up_package_method(my_perl, base_package, name);
-            SV * const rv = sv_2mortal(newRV((SV*)GvCV(gv)));
-            *count = call_sv(rv, flags);
-        }
+        GV * const gv = p5_look_up_package_method(my_perl, base_package, name);
+        SV * const rv = sv_2mortal(newRV_inc((SV*)GvCV(gv)));
+        *count = call_sv(rv, flags);
 
         SPAGAIN;
 
@@ -770,7 +784,7 @@ SV *p5_call_parent_gv(PerlInterpreter *my_perl, GV *gv, int len, SV *args[], I32
 
         PUTBACK;
 
-        SV * const rv = sv_2mortal(newRV((SV*)GvCV(gv))); /* FIXME: can be done once */
+        SV * const rv = GvCV(gv) ? sv_2mortal(newRV((SV*)GvCV(gv))) : (SV*)gv; /* FIXME: can be done once */
 
         *count = call_sv(rv, G_ARRAY | G_EVAL);
         SPAGAIN;
@@ -886,7 +900,8 @@ SV *p5_call_gv_two_args(PerlInterpreter *my_perl, GV *gv, SV *arg, SV *arg2, I32
         PUSHMARK(SP);
 
         XPUSHs(sv_2mortal((SV*)arg));
-        XPUSHs(sv_2mortal((SV*)arg2));
+        if (arg2 != NULL)
+            XPUSHs(sv_2mortal((SV*)arg2));
 
         PUTBACK;
 
@@ -923,7 +938,8 @@ SV *p5_scalar_call_gv_two_args(PerlInterpreter *my_perl, GV *gv, SV *arg, SV *ar
         PUSHMARK(SP);
 
         XPUSHs(sv_2mortal((SV*)arg));
-        XPUSHs(sv_2mortal((SV*)arg2));
+        if (arg2 != NULL)
+            XPUSHs(sv_2mortal((SV*)arg2));
 
         PUTBACK;
 
@@ -1262,6 +1278,8 @@ SV *p5_wrap_p6_callable(PerlInterpreter *my_perl, IV i, SV *p5obj) {
         SAVETMPS;
 
         PUSHMARK(SP);
+        XPUSHs(sv_2mortal(newSViv(i)));
+        PUTBACK;
         call_pv("Perl6::Callable::new", G_SCALAR);
         SPAGAIN;
 
@@ -1316,8 +1334,8 @@ SV *p5_wrap_p6_hash(
 
         PUSHMARK(SP);
 
-        XPUSHs(newSVpv("Perl6::Hash", 0));
-        XPUSHs(inst_ptr);
+        XPUSHs(sv_2mortal(newSVpv("Perl6::Hash", 0)));
+        XPUSHs(sv_2mortal(inst_ptr));
 
         PUTBACK;
 
@@ -1347,8 +1365,8 @@ SV *p5_wrap_p6_handle(PerlInterpreter *my_perl, IV i, SV *p5obj) {
 
         PUSHMARK(SP);
 
-        XPUSHs(newSVpv("Perl6::Handle", 0));
-        XPUSHs(handle);
+        XPUSHs(sv_2mortal(newSVpv("Perl6::Handle", 0)));
+        XPUSHs(sv_2mortal(handle));
 
         PUTBACK;
 
@@ -1507,7 +1525,7 @@ XS(p5_destroy_p5_object) {
         }
     }
     sp -= items;
-    XPUSHs(newSViv(destroyed && !PL_in_clean_objs));
+    XPUSHs(sv_2mortal(newSViv(destroyed && !PL_in_clean_objs)));
     XSRETURN(1);
 }
 
@@ -1614,20 +1632,15 @@ XS(p5_hash_assign_key) {
 
 XS(p5_call_p6_callable) {
     dXSARGS;
-    SV * obj = ST(0);
+    SV * index = ST(0);
 
     AV *args = create_args_array(ax, items, 1);
 
-    if (!SvROK(obj))
-        croak("Tried to call a Perl 6 method on a non-object!?");
-    SV * const obj_deref = SvRV(obj);
-    MAGIC * const mg = mg_find(obj_deref, '~');
-    _perl6_magic* const p6mg = (_perl6_magic*)(mg->mg_ptr);
     SV *err = NULL;
     SV * const args_rv = newRV_noinc((SV *) args);
 
     declare_cbs;
-    SV * retval = cbs->call_p6_callable(p6mg->index, args_rv, &err);
+    SV * retval = cbs->call_p6_callable(SvIV(index), args_rv, &err);
     return post_callback(ax, sp, items, args_rv, err, retval);
 }
 
